#include <stdlib.h>
#include <string.h>
#include <new>

#include "memory.h"
#include "common.h"
#include "error.h"
#include "osd.h"


// ubN蓖

// RomB[0]  : EmptyRom
// RomB[1]  : EmptyRam
// RomB[2]  : ExtRom
// RomB[3]  : ExtRom+0x2000
// RomB[4]  : MainRom
// RomB[5]  : MainRom+0x2000
// RomB[6]  : MainRom+0x4000
// RomB[7]  : MainRom+0x6000
// RomB[8]  : CGRom1
// RomB[9]  : CGRom2
// RomB[10] : KanjiRom
// RomB[11] : KanjiRom+0x2000
// RomB[12] : KanjiRom+0x4000
// RomB[13] : KanjiRom+0x6000
// RomB[14] : VoiceRom
// RomB[15] : VoiceRom+0x2000

// RamB[0]  : IntRam
// RamB[1]  : IntRam+0x2000
// RamB[2]  : IntRam+0x4000
// RamB[3]  : IntRam+0x6000
// RamB[4]  : IntRam+0x8000
// RamB[5]  : IntRam+0xa000
// RamB[6]  : IntRam+0xc000
// RamB[7]  : IntRam+0xe000
// RamB[8]  : ExtRam
// RamB[9]  : ExtRam+0x2000
// RamB[10] : ExtRam+0x4000
// RamB[11] : ExtRam+0x6000
// RamB[12] : ExtRam+0x8000
// RamB[13] : ExtRam+0xa000
// RamB[14] : ExtRam+0xc000
// RamB[15] : ExtRam+0xe000

#define EMPTYROM	(RomB[0])
#define EMPTYRAM	(RomB[1])
#define EXTROM0		(RomB[2])
#define EXTROM1		(RomB[3])
#define MAINROM0	(RomB[4])
#define MAINROM1	(RomB[5])
#define MAINROM2	(RomB[6])
#define MAINROM3	(RomB[7])
#define CGROM1		(RomB[8])
#define CGROM2		(RomB[9])
#define KANJIROM0	(RomB[10])
#define KANJIROM1	(RomB[11])
#define KANJIROM2	(RomB[12])
#define KANJIROM3	(RomB[13])
#define VOICEROM0	(RomB[14])
#define VOICEROM1	(RomB[15])
#define MAINRAM0	(RamB[0])
#define MAINRAM1	(RamB[1])
#define MAINRAM2	(RamB[2])
#define MAINRAM3	(RamB[3])
#define MAINRAM4	(RamB[4])
#define MAINRAM5	(RamB[5])
#define MAINRAM6	(RamB[6])
#define MAINRAM7	(RamB[7])
#define EXTRAM0		(RamB[8])
#define EXTRAM1		(RamB[9])
#define EXTRAM2		(RamB[10])
#define EXTRAM3		(RamB[11])
#define EXTRAM4		(RamB[12])
#define EXTRAM5		(RamB[13])
#define EXTRAM6		(RamB[14])
#define EXTRAM7		(RamB[15])
#define SOLRAM		(RamB[11])


#define pEMPTYROM	((MemBlock *)&EMPTYROM)
#define pEMPTYRAM	((MemBlock *)&EMPTYRAM)
#define pEXTROM0	(UseExtRom ? (MemBlock *)&EXTROM0 : pEMPTYROM)
#define pEXTROM1	(UseExtRom ? (MemBlock *)&EXTROM1 : pEMPTYROM)
#define pMAINROM0	((MemBlock *)&MAINROM0)
#define pMAINROM1	((MemBlock *)&MAINROM1)
#define pMAINROM2	((MemBlock *)&MAINROM2)
#define pMAINROM3	((MemBlock *)&MAINROM3)
#define pCGROM1		((MemBlock *)&CGROM1)
#define pCGROM2		((MemBlock *)&CGROM2)
#define pKANJIROM0	((MemBlock *)&KANJIROM0)
#define pKANJIROM1	((MemBlock *)&KANJIROM1)
#define pKANJIROM2	((MemBlock *)&KANJIROM2)
#define pKANJIROM3	((MemBlock *)&KANJIROM3)
#define pVOICEROM0	((MemBlock *)&VOICEROM0)
#define pVOICEROM1	((MemBlock *)&VOICEROM1)
#define pMAINRAM0	((MemBlock *)&MAINRAM0)
#define pMAINRAM1	((MemBlock *)&MAINRAM1)
#define pMAINRAM2	((MemBlock *)&MAINRAM2)
#define pMAINRAM3	((MemBlock *)&MAINRAM3)
#define pMAINRAM4	((MemBlock *)&MAINRAM4)
#define pMAINRAM5	((MemBlock *)&MAINRAM5)
#define pMAINRAM6	((MemBlock *)&MAINRAM6)
#define pMAINRAM7	((MemBlock *)&MAINRAM7)
#define pEXTRAM0	(UseExtRam ? (MemBlock *)&EXTRAM0 : pEMPTYRAM)
#define pEXTRAM1	(UseExtRam ? (MemBlock *)&EXTRAM1 : pEMPTYRAM)
#define pEXTRAM2	(UseExtRam ? (MemBlock *)&EXTRAM2 : pEMPTYRAM)
#define pEXTRAM3	(UseExtRam ? (MemBlock *)&EXTRAM3 : pEMPTYRAM)
#define pEXTRAM4	(UseExtRam ? (MemBlock *)&EXTRAM4 : pEMPTYRAM)
#define pEXTRAM5	(UseExtRam ? (MemBlock *)&EXTRAM5 : pEMPTYRAM)
#define pEXTRAM6	(UseExtRam ? (MemBlock *)&EXTRAM6 : pEMPTYRAM)
#define pEXTRAM7	(UseExtRam ? (MemBlock *)&EXTRAM7 : pEMPTYRAM)
#define pKNJROM0	(kj_rom ? ( kj_LR ? pKANJIROM2 : pKANJIROM0 ) : pVOICEROM0 )
#define pKNJROM1	(kj_rom ? ( kj_LR ? pKANJIROM3 : pKANJIROM1 ) : pVOICEROM1 )
#define pSOLRAM		((MemBlock *)&SOLRAM)


// Rg[WX^l
#define	INIT_RF0	(0x71)
#define INIT_RF1	(0xdd)
#define INIT_RF2	(0x50)


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
// 
//                                      ROMt@C  TCY   CRC32       l Wait(Read/Write)
const MEM6::MEMINFO MEM6::IEMPTROM  = { NULL,          0x02000, 0,          0xff,  1, 1 };
const MEM6::MEMINFO MEM6::IEMPTRAM  = { NULL,          0x02000, 0,          0xff,  0, 0 };
const MEM6::MEMINFO MEM6::IEXTROM   = { NULL,          0x20000, 0,          0xff,  1, 1 };
const MEM6::MEMINFO MEM6::IEXTRAM   = { NULL,          0x20000, 0,          0,     0, 0 };
										// MEGA-ROM/RAMJ[gbWΉ̂
										// gROM,RAM̈128KBmۂĂ

// m̃J[gbW ---------------------------------------------------------------------
const MEM6::MEMINFO MEM6::ISOLRAM   = { NULL,          0x02000, 0,          0xff,  1, 1 };
// ----------------------------------------------------------------------------------------

// PC-6001
const MEM6::MEMINFO MEM60::IMAINROM = { "BASICROM.60", 0x04000, 0x54c03109, 0xff,  1, 1 };
const MEM6::MEMINFO MEM60::ICGROM1  = { "CGROM60.60",  0x01000, 0xb0142d32, 0xff,  1, 1 };
const MEM6::MEMINFO MEM60::IINTRAM  = { NULL,          0x04000, 0,          0,     0, 0 };

// PC-6001mk2
const MEM6::MEMINFO MEM62::IMAINROM = { "BASICROM.62", 0x08000, 0x950ac401, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::ICGROM1  = { "CGROM60.62",  0x02000, 0x81eb5d95, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::ICGROM2  = { "CGROM60m.62", 0x02000, 0x3ce48c33, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IKANJI   = { "KANJIROM.62", 0x08000, 0x20c8f3eb, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IVOICE   = { "VOICEROM.62", 0x04000, 0x49b4f917, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IINTRAM  = { NULL,          0x10000, 0,          0,     0, 0 };

// PC-6601
const MEM6::MEMINFO MEM66::IMAINROM = { "BASICROM.66", 0x08000, 0xc0b01772, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::ICGROM1  = { "CGROM60.66",  0x02000, 0xd2434f29, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::ICGROM2  = { "CGROM66.66",  0x02000, 0x3ce48c33, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IKANJI   = { "KANJIROM.66", 0x08000, 0x20c8f3eb, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IVOICE   = { "VOICEROM.66", 0x04000, 0x91d078c1, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IINTRAM  = { NULL,          0x10000, 0,          0,     0, 0 };

// PC-6001mk2SR
// PC-6601SR



//--------------------------------------------------------------
// ubN
//--------------------------------------------------------------

////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
MemBlock::MemBlock( void )
{
	*Name = '\0';
	MB    = NULL;
	RWait = WWait = 0;
	WPt   = FALSE;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
MemBlock::~MemBlock( void ){}


////////////////////////////////////////////////////////////////
// ubNݒ
//
// :	name	ubN
//			mem		ubNւ̃|C^
//			rw		ANZXEFCg(Ǎ)
//			ww		ANZXEFCg()
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetMemory( const char *name, BYTE *mem, int rw, int ww )
{
	strncpy( Name, name, 32 );
	MB    = mem;
	RWait = rw;
	WWait = ww;
}


////////////////////////////////////////////////////////////////
// ANZXEFCgݒ
//
// :	rw		ANZXEFCg(Ǎ)
//			ww		ANZXEFCg()
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetWait( int rw, int ww )
{
	RWait = rw;
	WWait = ww;
}


////////////////////////////////////////////////////////////////
// ANZXEFCg擾
//
// :	Ȃ
// Ԓl:	bit0-7	ǍݎEFCg
//			bit8-15	ݎEFCg
////////////////////////////////////////////////////////////////
int MemBlock::GetWait( void )
{
	return ((WWait&0xff)<<8) | (RWait&0xff);
}


////////////////////////////////////////////////////////////////
// CgveNgݒ
//
// :	prt		CgveNgtO TRUE:Zbg FALSEF
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetProtect( BOOL prt )
{
	WPt = prt;
}


////////////////////////////////////////////////////////////////
// CgveNg擾
//
// :	Ȃ
// Ԓl:	BOOL	TRUE:Zbg FALSEF
////////////////////////////////////////////////////////////////
BOOL MemBlock::GetProtect( void )
{
	return WPt;
}


////////////////////////////////////////////////////////////////
// [h
////////////////////////////////////////////////////////////////
BYTE MemBlock::Read( WORD addr, int *wcnt )
{
	if( wcnt ) (*wcnt) += RWait;
	return MB[addr&0x1fff];
}


////////////////////////////////////////////////////////////////
// Cg
////////////////////////////////////////////////////////////////
void MemBlock::Write( WORD addr, BYTE data, int *wcnt )
{
	if( wcnt ) (*wcnt) += WWait;
	if( !WPt ) MB[addr&0x1fff] = data;
}




////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
MEM6::MEM6( BOOL extram )
{
	*FilePath    = '\0';
	
	UseExtRam    = extram;	// gRAMg?
	
	CGBank       = FALSE;	// CG ROM BANK 
	UseExtRom    = FALSE;	// gROM
	EnableChkCRC = TRUE;	// CRC`FbNL
	M1Wait       = 1;		// M1EFCgݒ
	
	memset( ram_sel, 0xff, sizeof(ram_sel) );
	
	// m̃J[gbW --------------------------
	UseSol  = FALSE;		// 
	SolBank = 0;			// ROMoN
	// ----------------------------------------------
	
}

MEM60::MEM60( const ID& id, BOOL extram ) : MEM6(extram), Device(id){}

MEM62::MEM62( const ID& id, BOOL extram ) : MEM6(extram), Device(id)
{
	Rf0 = INIT_RF0;			// Rg[WX^lݒ
	Rf1 = INIT_RF1;			// Rg[WX^lݒ
	Rf2 = INIT_RF2;			// Rg[WX^lݒ
	
	cgrom  = TRUE;
	kj_rom = TRUE;
	kj_LR  = TRUE;
}

MEM66::MEM66( const ID& id, BOOL extram ) : MEM62(id,extram){}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
MEM6::~MEM6( void )
{
	if( MainRom )	delete[] MainRom;
	if( IntRam )	delete[] IntRam;
	if( CGRom1 )	delete[] CGRom1;
	if( ExtRom )	delete[] ExtRom;
	if( EmptyRom )	delete[] EmptyRom;
	if( EmptyRam )	delete[] EmptyRam;
}

MEM60::~MEM60( void ){}

MEM62::~MEM62( void )
{
	if( CGRom2 )	delete[] CGRom2;
	if( KanjiRom )	delete[] KanjiRom;
	if( VoiceRom )	delete[] VoiceRom;
}

MEM66::~MEM66( void ){}


////////////////////////////////////////////////////////////////
// gROM }Eg
////////////////////////////////////////////////////////////////
BOOL MEM6::MountExtRom( char *filename )
{
	PRINTD1( MEM_LOG, "[MEM][MountExtRom] -> %s\n", filename );
	
	DWORD LoadSize = 0;
	
	UnmountExtRom();	// UJ
	
	try{
		FILE *fp = FOPENEN( filename, "rb" );
		if( !fp ) throw Error::ExtRomMountFailed;
		LoadSize = fread( ExtRom, sizeof(BYTE), IEXTROM.Size, fp );
		fclose( fp );
		
		// TCY`FbN
//		if( LoadSize != IEXTROM.Size ) throw Error::RomSizeNG;
		
		// t@CpXۑ
		strncpy( FilePath, filename, PATH_MAX );
	}
	// O
	catch( Error::Errno i ){
		Error::SetError( i );
		
		UnmountExtRom();
		return FALSE;
	}
	
	UseExtRom = TRUE;
	Reset();
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// gROM A}Eg
////////////////////////////////////////////////////////////////
void MEM6::UnmountExtRom( void )
{
	PRINTD( MEM_LOG, "[MEM][UnmountExtRom]\n" );
	
	memset( ExtRom, IEXTROM.Init, IEXTROM.Size );
	*FilePath = '\0';
	
	UseExtRom = FALSE;
	Reset();
}


////////////////////////////////////////////////////////////////
// gROMt@CpX擾
////////////////////////////////////////////////////////////////
char *MEM6::GetFile( void )
{
	return FilePath;
}


////////////////////////////////////////////////////////////////
// mۂROMt@CǍ
////////////////////////////////////////////////////////////////
BOOL MEM6::AllocMemory( BYTE **buf, const MEMINFO *info, char *path )
{
	PRINTD1( MEM_LOG, "[MEM][AllocMemory] -> %s\n", info->FileName );
	
	BYTE *buffer = NULL;
	DWORD LoadSize = 0;
	
	try{
		// m(ŏTCY8KB)
		if( info->Size < 0x2000 ) buffer = new BYTE[0x2000];
		else                      buffer = new BYTE[info->Size];
		memset( buffer, info->Init, info->Size );
		
		*buf = buffer;
		
		// t@C=NULL ȂRAM܂NULL
		if( !info->FileName ) return TRUE;
		
		// t@CǍ
		char fpath[PATH_MAX] = "";
		if( path ){
			strncpy( fpath, path, PATH_MAX );
			AddDelimiter( fpath );
		}
		strncat( fpath, info->FileName, PATH_MAX );
		
		FILE *fp = FOPENEN( fpath, "rb" );
		if( !fp ) throw Error::NoRom;
		LoadSize = fread( buffer, sizeof(BYTE), info->Size, fp );
		fclose( fp );
		
		// TCY`FbN
		if( LoadSize != info->Size ) throw Error::RomSizeNG;
		
		// CRC`FbN
		if( EnableChkCRC && (info->Crc != 0) ){	// EnableChkCRC=FALSE ܂ CRC=0̎̓`FbNȂ
			if( CalcCrc32( buffer, info->Size ) !=info->Crc ) throw Error::RomCrcNG;
		}
	}
	// new Ɏsꍇ
	catch( std::bad_alloc ){
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	// O
	catch( Error::Errno i ){
		Error::SetError( i );
		
		switch(i){
		case Error::NoRom:		// t@C̃I[vɎsꍇ
		case Error::RomSizeNG:	// TCYȂꍇ
		case Error::RomCrcNG:	// CRCȂꍇ
		default:				// J
			delete [] buffer;
			*buf = NULL;
		}
		return FALSE;
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL MEM6::InitCommon( BOOL chkcrc, BOOL usesol )
{
	PRINTD( MEM_LOG, "[MEM][InitCommon]\n" );
	
	EnableChkCRC = chkcrc;
	UseSol       = usesol;	// m̃J[gbWL
	
	// m̃J[gbWgꍇ͕KORAML
	if( UseSol ) UseExtRam = TRUE;
	
	// m
	if( !AllocMemory( &EmptyRom, &IEMPTROM, NULL ) ) return FALSE;
	if( !AllocMemory( &EmptyRam, &IEMPTRAM, NULL ) ) return FALSE;
	
	if( !AllocMemory( &ExtRom,   &IEXTROM,  NULL ) ) return FALSE;
	if( !AllocMemory( &ExtRam,   &IEXTRAM,  NULL ) ) return FALSE;
	
	
	// ubNݒ
	// Ƃ肠SEmptyɐݒ
	for( int i=0; i<MAXRMB; i++ ){
		RomB[i].SetMemory( "EmptyRom", EmptyRom, IEMPTROM.WaitR, IEMPTROM.WaitW );
		RomB[i].SetProtect( TRUE );		// CgveNgL
	}
	for( int i=0; i<MAXWMB; i++ ){
		RamB[i].SetMemory( "EmptyRam", EmptyRam, IEMPTRAM.WaitR, IEMPTRAM.WaitW );
		RomB[i].SetProtect( FALSE );	// CgveNg
	}
	
	// EmptyRom,EmptyRam
	//   EmptyRam(߂Ȃ̂)EmptyRomƓ
	//   ANZXEFCĝ݈قȂ
	EMPTYROM.SetMemory( "EmptyRom", EmptyRom, IEMPTROM.WaitR, IEMPTROM.WaitW );
	EMPTYRAM.SetMemory( "EmptyRam", EmptyRom, IEMPTRAM.WaitR, IEMPTRAM.WaitW );
	EMPTYRAM.SetProtect( TRUE );		// CgveNgL
	
	// gROM̈
	EXTROM0.SetMemory( "ExtRom0", ExtRom,        IEXTROM.WaitR, IEXTROM.WaitW );
	EXTROM1.SetMemory( "ExtRom1", ExtRom+0x2000, IEXTROM.WaitR, IEXTROM.WaitW );
	
	// ORAM
	if( UseExtRam ){
		EXTRAM0.SetMemory( "ExtRam0", ExtRam,        IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM1.SetMemory( "ExtRam1", ExtRam+0x2000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM2.SetMemory( "ExtRam2", ExtRam+0x4000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM3.SetMemory( "ExtRam3", ExtRam+0x6000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM4.SetMemory( "ExtRam4", ExtRam+0x8000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM5.SetMemory( "ExtRam5", ExtRam+0xa000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM6.SetMemory( "ExtRam6", ExtRam+0xc000, IEXTRAM.WaitR, IEXTRAM.WaitW );
		EXTRAM7.SetMemory( "ExtRam7", ExtRam+0xe000, IEXTRAM.WaitR, IEXTRAM.WaitW );
	}
	
	// m̃J[gbWRAM̈
	if( UseSol )	// ORAM̈ꕔ𗘗p
		SOLRAM.SetMemory( "SolRam", ExtRam+0x6000, ISOLRAM.WaitR, ISOLRAM.WaitW );
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] ExtRom   :%08X\n", (int)ExtRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] ExtRam   :%08X\n", (int)ExtRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] EmptyRom :%08X\n", (int)EmptyRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] EmptyRam :%08X\n", (int)EmptyRam );
	
	return TRUE;
}

BOOL MEM60::Init( char *path, BOOL chkcrc, BOOL usesol )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc, usesol ) ) return FALSE;	// ʕ
	
	// m
	if( !AllocMemory( &MainRom, &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,  &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &IntRam,  &IINTRAM,  path ) ) return FALSE;
	
	
	// ubNݒ
	// BASIC ROM
	MAINROM0.SetMemory( "MainRom0", MainRom,        IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM1.SetMemory( "MainRom1", MainRom+0x2000, IMAINROM.WaitR, IMAINROM.WaitW );
	
	// CG ROM
	CGROM1.SetMemory( "CGRom1", CGRom1, ICGROM1.WaitR, ICGROM1.WaitW );
	
	// RAM
	MAINRAM0.SetMemory( "IntRam0", IntRam,        IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM1.SetMemory( "IntRam1", IntRam+0x2000, IINTRAM.WaitR, IINTRAM.WaitW );
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] IntRam  :%08X\n", (int)IntRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1  :%08X\n", (int)CGRom1 );
	
	return TRUE;
}

BOOL MEM62::Init( char *path, BOOL chkcrc, BOOL usesol )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc, usesol ) ) return FALSE;	// ʕ
	
	// m
	if( !AllocMemory( &MainRom,  &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,   &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &CGRom2,   &ICGROM2,  path ) ) return FALSE;
	if( !AllocMemory( &KanjiRom, &IKANJI,   path ) ) return FALSE;
	if( !AllocMemory( &VoiceRom, &IVOICE,   path ) ) return FALSE;
	if( !AllocMemory( &IntRam,   &IINTRAM,  path ) ) return FALSE;
	
	
	// ubNݒ
	// BASIC ROM
	MAINROM0.SetMemory( "MainRom0", MainRom,        IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM1.SetMemory( "MainRom1", MainRom+0x2000, IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM2.SetMemory( "MainRom2", MainRom+0x4000, IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM3.SetMemory( "MainRom3", MainRom+0x6000, IMAINROM.WaitR, IMAINROM.WaitW );
	
	// CG ROM
	CGROM1.SetMemory( "CGRom1", CGRom1, ICGROM1.WaitR, ICGROM1.WaitW );
	CGROM2.SetMemory( "CGRom2", CGRom2, ICGROM2.WaitR, ICGROM2.WaitW );
	
	// ROM
	KANJIROM0.SetMemory( "KanjiRom0", KanjiRom,        IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM1.SetMemory( "KanjiRom1", KanjiRom+0x2000, IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM2.SetMemory( "KanjiRom2", KanjiRom+0x4000, IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM3.SetMemory( "KanjiRom3", KanjiRom+0x6000, IKANJI.WaitR, IKANJI.WaitW );
	
	// ROM
	VOICEROM0.SetMemory( "VoiceRom0", VoiceRom,        IVOICE.WaitR, IVOICE.WaitW );
	VOICEROM1.SetMemory( "VoiceRom1", VoiceRom+0x2000, IVOICE.WaitR, IVOICE.WaitW );
	
	// RAM
	MAINRAM0.SetMemory( "IntRam0", IntRam,        IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM1.SetMemory( "IntRam1", IntRam+0x2000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM2.SetMemory( "IntRam2", IntRam+0x4000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM3.SetMemory( "IntRam3", IntRam+0x6000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM4.SetMemory( "IntRam4", IntRam+0x8000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM5.SetMemory( "IntRam5", IntRam+0xa000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM6.SetMemory( "IntRam6", IntRam+0xc000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM7.SetMemory( "IntRam7", IntRam+0xe000, IINTRAM.WaitR, IINTRAM.WaitW );
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] IntRam   :%08X\n", (int)IntRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom  :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1   :%08X\n", (int)CGRom1 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom2   :%08X\n", (int)CGRom2 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] KanjiRom :%08X\n", (int)KanjiRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] VoiceRom :%08X\n", (int)VoiceRom );
	
	return TRUE;
}

BOOL MEM66::Init( char *path, BOOL chkcrc, BOOL usesol )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc, usesol ) ) return FALSE;	// ʕ
	
	// m
	if( !AllocMemory( &MainRom,  &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,   &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &CGRom2,   &ICGROM2,  path ) ) return FALSE;
	if( !AllocMemory( &KanjiRom, &IKANJI,   path ) ) return FALSE;
	if( !AllocMemory( &VoiceRom, &IVOICE,   path ) ) return FALSE;
	if( !AllocMemory( &IntRam,  &IINTRAM,  path ) ) return FALSE;
	
	
	// ubNݒ
	// BASIC ROM
	MAINROM0.SetMemory( "MainRom0", MainRom,        IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM1.SetMemory( "MainRom1", MainRom+0x2000, IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM2.SetMemory( "MainRom2", MainRom+0x4000, IMAINROM.WaitR, IMAINROM.WaitW );
	MAINROM3.SetMemory( "MainRom3", MainRom+0x6000, IMAINROM.WaitR, IMAINROM.WaitW );
	
	// CG ROM
	CGROM1.SetMemory( "CGRom1", CGRom1, ICGROM1.WaitR, ICGROM1.WaitW );
	CGROM2.SetMemory( "CGRom2", CGRom2, ICGROM2.WaitR, ICGROM2.WaitW );
	
	// ROM
	KANJIROM0.SetMemory( "KanjiRom0", KanjiRom,        IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM1.SetMemory( "KanjiRom1", KanjiRom+0x2000, IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM2.SetMemory( "KanjiRom2", KanjiRom+0x4000, IKANJI.WaitR, IKANJI.WaitW );
	KANJIROM3.SetMemory( "KanjiRom3", KanjiRom+0x6000, IKANJI.WaitR, IKANJI.WaitW );
	
	// ROM
	VOICEROM0.SetMemory( "VoiceRom0", VoiceRom,        IVOICE.WaitR, IVOICE.WaitW );
	VOICEROM1.SetMemory( "VoiceRom1", VoiceRom+0x2000, IVOICE.WaitR, IVOICE.WaitW );
	
	// RAM
	MAINRAM0.SetMemory( "IntRam0", IntRam,        IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM1.SetMemory( "IntRam1", IntRam+0x2000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM2.SetMemory( "IntRam2", IntRam+0x4000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM3.SetMemory( "IntRam3", IntRam+0x6000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM4.SetMemory( "IntRam4", IntRam+0x8000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM5.SetMemory( "IntRam5", IntRam+0xa000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM6.SetMemory( "IntRam6", IntRam+0xc000, IINTRAM.WaitR, IINTRAM.WaitW );
	MAINRAM7.SetMemory( "IntRam7", IntRam+0xe000, IINTRAM.WaitR, IINTRAM.WaitW );
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] IntRam   :%08X\n", (int)IntRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom  :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1   :%08X\n", (int)CGRom1 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom2   :%08X\n", (int)CGRom2 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] KanjiRom :%08X\n", (int)KanjiRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] VoiceRom :%08X\n", (int)VoiceRom );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// pb`
////////////////////////////////////////////////////////////////
void MEM60::Patch( void )
{
	// WCXeBbN̓[` oOC(sS)
	MainRom[0x1cb4] = 0xe6;	// AND A,x
	MainRom[0x1cb5] = 0xbf;
}

void MEM62::Patch( void )
{
	// WCXeBbN̓[` oOC(sS)
	MainRom[0x1cb4] = 0xe6;	// AND A,x
	MainRom[0x1cb5] = 0xbf;
}

void MEM66::Patch( void )
{
	// WCXeBbN̓[` oOC(sS)
	MainRom[0x1cb4] = 0xe6;	// AND A,x
	MainRom[0x1cb5] = 0xbf;
}


////////////////////////////////////////////////////////////////
// Zbg
////////////////////////////////////////////////////////////////
void MEM60::Reset( void )
{
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( 0 );	// oN
	// -------------------------------------------------------------
	
	CGBank = FALSE;	// CG ROM BANK 
	
	// CGROM̌㔼4KBɊgROM̓eRs[
	// (oN؂ւ8KBPʂōs)
	// gROMꍇCGROM̃C[WƂȂ炵
	memcpy( &CGRom1[0x1000], UseExtRom ? &ExtRom[0x3000] : &CGRom1[0x0000], 0x1000 );
	
	SetMB_60();
}

void MEM62::Reset( void )
{
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( 0 );	// oN
	// -------------------------------------------------------------
	
	CGBank = FALSE;	// CG ROM BANK 
	Rf0 = INIT_RF0;	// Rg[WX^lݒ
	Rf1 = INIT_RF1;	// Rg[WX^lݒ
	Rf2 = INIT_RF2;	// Rg[WX^lݒ
	
	SetMB_R( Rf0, Rf1 );
	SetMB_W( Rf2 );
}


////////////////////////////////////////////////////////////////
// tFb`(M1)
////////////////////////////////////////////////////////////////
BYTE MEM6::Fetch( WORD addr, int *m1wait )
{
	PRINTD1( MEM_LOG, "[MEM][Fetch] -> %04X", addr );
	
	// M1EFCgǉ
	if( m1wait ) (*m1wait) += M1Wait;
	
	return Rm_blk[addr>>13]->Read( addr );
}


////////////////////////////////////////////////////////////////
// [h
////////////////////////////////////////////////////////////////
BYTE MEM6::Read( WORD addr, int *wcnt )
{
	PRINTD1( MEM_LOG, "[MEM][Read] -> %04X", addr );
	
	return Rm_blk[addr>>13]->Read( addr, wcnt );
}


////////////////////////////////////////////////////////////////
// Cg
////////////////////////////////////////////////////////////////
void MEM6::Write( WORD addr, BYTE data, int *wcnt )
{
	PRINTD5( MEM_LOG, "[MEM][Write] %04X:%02X -> %08X[%04X]'%c'", addr, data, (int)Wm_blk[addr>>13], addr&0x1fff, data );
	
	Wm_blk[addr>>13]->Write( addr, data, wcnt );
	
	// /ORAMƂ݂̏ꍇ͂ЂƂ܂
//	if( ram_sel[addr>>14]&3 == 3 ){
//	}
}


////////////////////////////////////////////////////////////////
// [h̃ubNw(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetMB_R( BYTE mem1, BYTE mem2 )
{
	PRINTD( MEM_LOG, "[MEM][SetMB_R]\n" );
	
	// Port F0H
	switch( mem1 & 0x0f ){	// RF0 (0x0000 - 0x3fff)
		case 0x00:	Rm_blk[0] = pEMPTYROM;	Rm_blk[1] = pEMPTYROM;	break;
		case 0x01:	Rm_blk[0] = pMAINROM0;	Rm_blk[1] = pMAINROM1;	break;
		case 0x02:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x03:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = pEXTROM1;	break;
		case 0x04:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pEXTROM0;	break;
		case 0x05:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = pMAINROM1;	break;
		case 0x06:	Rm_blk[0] = pMAINROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x07:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pEXTROM1;	break;
		case 0x08:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = pEXTROM0;	break;
		case 0x09:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = pMAINROM1;	break;
		case 0x0a:	Rm_blk[0] = pMAINROM0;	Rm_blk[1] = pEXTROM1;	break;
		case 0x0b:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x0c:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = pEXTROM0;	break;
		case 0x0d:	Rm_blk[0] = pMAINRAM0;	Rm_blk[1] = pMAINRAM1;	break;
		case 0x0e:	Rm_blk[0] = pEXTRAM0;	Rm_blk[1] = pEXTRAM1;	break;
		case 0x0f:	Rm_blk[0] = pEMPTYROM;	Rm_blk[1] = pEMPTYROM;	break;
	}
	switch( mem1 & 0xf0 ){	// RF0 (0x4000 - 0x7fff)
		case 0x00:	Rm_blk[2] = pEMPTYROM;	Rm_blk[3] = pEMPTYROM;	break;
		case 0x10:	Rm_blk[2] = pMAINROM2;	Rm_blk[3] = pMAINROM3;	break;
		case 0x20:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = pKNJROM1;	break;
		case 0x30:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = UseSol ? pSOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0x40:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = pEXTROM0;	break;
		case 0x50:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = pMAINROM3;	break;
		case 0x60:	Rm_blk[2] = pMAINROM2;	Rm_blk[3] = pKNJROM1;	break;
		case 0x70:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = UseSol ? pSOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0x80:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = pEXTROM0;	break;
		case 0x90:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = pMAINROM3;	break;
		case 0xa0:	Rm_blk[2] = pMAINROM2;	Rm_blk[3] = UseSol ? pSOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0xb0:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = pKNJROM1;	break;
		case 0xc0:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = pEXTROM0;	break;
		case 0xd0:	Rm_blk[2] = pMAINRAM2;	Rm_blk[3] = pMAINRAM3;	break;
		case 0xe0:	Rm_blk[2] = pEXTRAM2;	Rm_blk[3] = pEXTRAM3;	break;
		case 0xf0:	Rm_blk[2] = pEMPTYROM;	Rm_blk[3] = pEMPTYROM;	break;
	}
	if( CGBank ) Rm_blk[3] = cgrom ? pCGROM1 : pCGROM2;
	
	// Port F1H
	switch( mem2 & 0x0f ){	// RF1 (0x8000 - 0xbfff)
		case 0x00:	Rm_blk[4] = pEMPTYROM;	Rm_blk[5] = pEMPTYROM;	break;
		case 0x01:	Rm_blk[4] = pMAINROM0;	Rm_blk[5] = pMAINROM1;	break;
		case 0x02:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x03:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = pEXTROM1;	break;
		case 0x04:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pEXTROM0;	break;
		case 0x05:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = pMAINROM1;	break;
		case 0x06:	Rm_blk[4] = pMAINROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x07:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pEXTROM1;	break;
		case 0x08:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = pEXTROM0;	break;
		case 0x09:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = pMAINROM1;	break;
		case 0x0a:	Rm_blk[4] = pMAINROM0;	Rm_blk[5] = pEXTROM1;	break;
		case 0x0b:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x0c:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = pEXTROM0;	break;
		case 0x0d:	Rm_blk[4] = pMAINRAM4;	Rm_blk[5] = pMAINRAM5;	break;
		case 0x0e:	Rm_blk[4] = pEXTRAM4;	Rm_blk[5] = pEXTRAM5;	break;
		case 0x0f:	Rm_blk[4] = pEMPTYROM;	Rm_blk[5] = pEMPTYROM;	break;
	}
	switch( mem2 & 0xf0 ){	// RF1 (0xc000 - 0xffff)
		case 0x00:	Rm_blk[6] = pEMPTYROM;	Rm_blk[7] = pEMPTYROM;	break;
		case 0x10:	Rm_blk[6] = pMAINROM2;	Rm_blk[7] = pMAINROM3;	break;
		case 0x20:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = pKNJROM1;	break;
		case 0x30:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = pEXTROM1;	break;
		case 0x40:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pEXTROM0;	break;
		case 0x50:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = pMAINROM3;	break;
		case 0x60:	Rm_blk[6] = pMAINROM2;	Rm_blk[7] = pKNJROM1;	break;
		case 0x70:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pEXTROM1;	break;
		case 0x80:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = pEXTROM0;	break;
		case 0x90:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = pMAINROM3;	break;
		case 0xa0:	Rm_blk[6] = pMAINROM2;	Rm_blk[7] = pEXTROM1;	break;
		case 0xb0:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pKNJROM1;	break;
		case 0xc0:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = pEXTROM0;	break;
		case 0xd0:	Rm_blk[6] = pMAINRAM6;	Rm_blk[7] = pMAINRAM7;	break;
		case 0xe0:	Rm_blk[6] = pEXTRAM6;	Rm_blk[7] = pEXTRAM7;	break;
		case 0xf0:	Rm_blk[6] = pEMPTYROM;	Rm_blk[7] = pEMPTYROM;	break;
	}
	
	// WX^ۑ
	Rf0 = mem1;
	Rf1 = mem2;
	
	#if (MEM_LOG)
//		char mb[32]; // fixed by Windy:  mb ͗vȂ炵B
		fprintf( stdout, "[MEM][SetMB_R] %02X %02X\n", mem1, mem2 );
//		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetReadMemBlk(i,mb) );//bug fix
		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetReadMemBlk(i) );
	#endif
}



////////////////////////////////////////////////////////////////
// Cg̃ubNw(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetMB_W( BYTE data )
{
	PRINTD1( MEM_LOG, "[MEM][SetMB_W] -> %02X\n", data );
	
	switch( ram_sel[0] = data & 3 ){
	case 0: Wm_blk[0] = pEMPTYRAM;	Wm_blk[1] = pEMPTYRAM;	break;
	case 1: Wm_blk[0] = pMAINRAM0;	Wm_blk[1] = pMAINRAM1;	break;
	case 2: Wm_blk[0] = pEXTRAM0;	Wm_blk[1] = pEXTRAM1;	break;
	case 3: Wm_blk[0] = pMAINRAM0;	Wm_blk[1] = pMAINRAM1;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( ram_sel[1] = (data>>2) & 3 ){
	case 0: Wm_blk[2] = pEMPTYRAM;	Wm_blk[3] = pEMPTYRAM;	break;
	case 1: Wm_blk[2] = pMAINRAM2;	Wm_blk[3] = UseSol ? pSOLRAM : pMAINRAM3;	break;	// m̃J[gbWΉ
	case 2: Wm_blk[2] = pEXTRAM2;	Wm_blk[3] = UseSol ? pSOLRAM : pEXTRAM3;	break;	// m̃J[gbWΉ
	case 3: Wm_blk[2] = pMAINRAM2;	Wm_blk[3] = UseSol ? pSOLRAM : pMAINRAM3;	break;	// m̃J[gbWΉ
																						// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( ram_sel[2] = (data>>4) & 3 ){
	case 0: Wm_blk[4] = pEMPTYRAM;	Wm_blk[5] = pEMPTYRAM;	break;
	case 1: Wm_blk[4] = pMAINRAM4;	Wm_blk[5] = pMAINRAM5;	break;
	case 2: Wm_blk[4] = pEXTRAM4;	Wm_blk[5] = pEXTRAM5;	break;
	case 3: Wm_blk[4] = pMAINRAM4;	Wm_blk[5] = pMAINRAM5;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( ram_sel[3] = (data>>6) & 3 ){
	case 0: Wm_blk[6] = pEMPTYRAM;	Wm_blk[7] = pEMPTYRAM;	break;
	case 1: Wm_blk[6] = pMAINRAM6;	Wm_blk[7] = pMAINRAM7;	break;
	case 2: Wm_blk[6] = pEXTRAM6;	Wm_blk[7] = pEXTRAM7;	break;
	case 3: Wm_blk[6] = pMAINRAM6;	Wm_blk[7] = pMAINRAM7;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	
	// WX^ۑ
	Rf2 = data;
	
	#if (MEM_LOG)
//		char mb[32]; // fixed by Windy  mb ͗vȂ炵B
//		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetWriteMemBlk(i,mb) );// bug fix
		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetWriteMemBlk(i) );
	#endif
}


////////////////////////////////////////////////////////////////
// ANZXEFCgݒ(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetWait( BYTE data )
{
	PRINTD3( MEM_LOG, "[MEM][SetWait] -> M1:%d ROM:%d RAM:%d\n", (data>>7)&1, (data>>6)&1, (data>>5)&1 );
	
	// M1
	M1Wait = data&0x80 ? 1 : 0;
	
	// ROM
	int romwait = data&0x40 ? 1 : 0;
	EMPTYROM.SetWait( romwait, romwait );
	EXTROM0.SetWait( romwait, romwait );
	EXTROM1.SetWait( romwait, romwait );
	MAINROM0.SetWait( romwait, romwait );
	MAINROM1.SetWait( romwait, romwait );
	MAINROM2.SetWait( romwait, romwait );
	MAINROM3.SetWait( romwait, romwait );
	KANJIROM0.SetWait( romwait, romwait );
	KANJIROM1.SetWait( romwait, romwait );
	KANJIROM2.SetWait( romwait, romwait );
	KANJIROM3.SetWait( romwait, romwait );
	VOICEROM0.SetWait( romwait, romwait );
	VOICEROM1.SetWait( romwait, romwait );
	
	// RAM
	int ramwait = data&0x20 ? 1 : 0;
	EMPTYRAM.SetWait( ramwait, ramwait );
	MAINRAM0.SetWait( ramwait, ramwait );
	MAINRAM1.SetWait( ramwait, ramwait );
	MAINRAM2.SetWait( ramwait, ramwait );
	MAINRAM3.SetWait( ramwait, ramwait );
	MAINRAM4.SetWait( ramwait, ramwait );
	MAINRAM5.SetWait( ramwait, ramwait );
	MAINRAM6.SetWait( ramwait, ramwait );
	MAINRAM7.SetWait( ramwait, ramwait );
	EXTRAM0.SetWait( ramwait, ramwait );
	EXTRAM1.SetWait( ramwait, ramwait );
	EXTRAM2.SetWait( ramwait, ramwait );
	EXTRAM3.SetWait( ramwait, ramwait );
	EXTRAM4.SetWait( ramwait, ramwait );
	EXTRAM5.SetWait( ramwait, ramwait );
	EXTRAM6.SetWait( ramwait, ramwait );
	EXTRAM7.SetWait( ramwait, ramwait );
}


////////////////////////////////////////////////////////////////
// ANZXEFCg擾(62,66̂)
////////////////////////////////////////////////////////////////
BYTE MEM62::GetWait( void )
{
	return ( M1Wait ? 0x80 : 0 ) | ( EMPTYROM.GetWait() ? 0x40 : 0 ) | ( EMPTYRAM.GetWait() ? 0x20 : 0 );
}


////////////////////////////////////////////////////////////////
// ubNݒ(60̂)
////////////////////////////////////////////////////////////////
void MEM60::SetMB_60( void )
{
	PRINTD( MEM_LOG, "[MEM][SetMB_60]\n" );
	
	// Read
	Rm_blk[0] = pMAINROM0; Rm_blk[1] = pMAINROM1;
	Rm_blk[2] = pEXTROM0;  Rm_blk[3] = UseSol ? pSOLRAM : pEXTROM1;	// m̃J[gbWg?
	Rm_blk[4] = pEXTRAM0;  Rm_blk[5] = pEXTRAM1;
	Rm_blk[6] = pMAINRAM0; Rm_blk[7] = pMAINRAM1;
	
	// Write
	Wm_blk[0] = pMAINROM0; Wm_blk[1] = pMAINROM1;
	Wm_blk[2] = pEXTROM0;  Wm_blk[3] = UseSol ? pSOLRAM : pEXTROM1; // m̃J[gbWg?
	Wm_blk[4] = pEXTRAM0;  Wm_blk[5] = pEXTRAM1;
	Wm_blk[6] = pMAINRAM0; Wm_blk[7] = pMAINRAM1;
	
	// ram_sel ̐ݒ͂Ԃsv
}


////////////////////////////////////////////////////////////////
// CG ROM I(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SelectCGrom( int mode )
{
	PRINTD1( MEM_LOG, "[MEM][SelectCGrom] -> %d\n", mode );
	
	// mode 1:32*16(N60[h) 0:40*20(N60m[h)
	cgrom = mode ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// CG ROM BANK ؂ւ
////////////////////////////////////////////////////////////////
void MEM6::SetCGBank( BOOL data )
{
	PRINTD1( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	Rm_blk[3] = CGBank ? pCGROM1 : UseSol ? (MemBlock*)pSOLRAM : pEXTROM1;
}

void MEM62::SetCGBank( BOOL data )
{
	PRINTD1( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	SetMB_R( Rf0, Rf1 );
}


////////////////////////////////////////////////////////////////
// ROM  ROM ؂ւ(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SelectKanjiRom( BYTE mode )
{
	PRINTD1( MEM_LOG, "[MEM][SelectKanjiRom] -> %02X\n", mode );
	
// mode bit0 0:ROMI 1:ROMI
//      bit1 0:ROM     1:ROME

	//  HEH
	kj_LR = mode&0x02 ? TRUE : FALSE;
	
	// ROMHROMH
	kj_rom = mode&0x01 ? TRUE : FALSE;
	
	SetMB_R( Rf0, Rf1 );
}

void MEM66::SelectKanjiRom( BYTE mode )
{
	MEM62::SelectKanjiRom( mode );
	
	// 66̏ꍇ,ƈႤ炵
	if( !(mode&0x01) ){
		if( Rm_blk[0] != pMAINROM0 ) Rm_blk[0] = pVOICEROM0;
		if( Rm_blk[1] != pMAINROM1 ) Rm_blk[1] = pVOICEROM1;
	}
}


////////////////////////////////////////////////////////////////
// m̃J[gbW ROMoNݒ
////////////////////////////////////////////////////////////////
void MEM6::SetSolBank( BYTE data )
{
	SolBank = data & 0xf;
	EXTROM0.SetMemory( EXTROM0.GetName(), ExtRom + 0x2000 * SolBank, IEXTROM.WaitR, IEXTROM.WaitW );
	
	// RAM̃CgveNg
	SOLRAM.SetProtect( data & 0x10 ? TRUE : FALSE );
}


////////////////////////////////////////////////////////////////
// ǂłSAVE
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	BOOL	TRUE: FALSE:s
////////////////////////////////////////////////////////////////
BOOL MEM60::DokoSave( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "MEMORY", NULL, "CGBank",	 "%s",	CGBank ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "UseExtRam",	 "%s",	UseExtRam ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "M1Wait",	 "%d",	M1Wait );
	Ini->PutEntry( "MEMORY", NULL, "UseSoldier", "%s",	UseSol ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "SolBank",	 "%d",	SolBank );
	
	// gROM}EgĂꍇ
	if( UseExtRom )
		Ini->PutEntry( "MEMORY", NULL, "FilePath",	"%s",	FilePath );
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", IntRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	// ORAM
	if( UseExtRam ){
		for( int i=0; i<0x4000; i+=64 ){
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
		}
		if( UseSol ){
			for( int i=0x6000; i<0x8000; i+=64 ){
				char stren[16],strva[256];
				sprintf( stren, "ExtRam_%04X", i );
				for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
				Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
			}
		}
	}
	
	return TRUE;
}

BOOL MEM62::DokoSave( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "MEMORY", NULL, "CGBank",	 "%s",	CGBank ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "UseExtRam",	 "%s",	UseExtRam ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "cgrom",		 "%s",	cgrom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_rom",	 "%s",	kj_rom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_LR",		 "%s",	kj_LR ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "Rf0",		 "%d",	Rf0 );
	Ini->PutEntry( "MEMORY", NULL, "Rf1",		 "%d",	Rf1 );
	Ini->PutEntry( "MEMORY", NULL, "Rf2",		 "%d",	Rf2 );
	Ini->PutEntry( "MEMORY", NULL, "M1Wait",	 "%d",	M1Wait );
	Ini->PutEntry( "MEMORY", NULL, "UseSoldier", "%s",	UseSol ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "SolBank",	 "%d",	SolBank );
	
	// gROM}EgĂꍇ
	if( UseExtRom )
		Ini->PutEntry( "MEMORY", NULL, "FilePath",	"%s",	FilePath );
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", IntRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	// ORAM
	if( UseExtRam ){
		for( int i=0; i<(int)IEXTRAM.Size; i+=64 ){
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
		}
	}
	
	return TRUE;
}



////////////////////////////////////////////////////////////////
// ǂłLOAD
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	BOOL	TRUE: FALSE:s
////////////////////////////////////////////////////////////////
BOOL MEM60::DokoLoad( cIni *Ini )
{
	char strfp[PATH_MAX];
	
	if( !Ini ) return FALSE;
	
	Ini->GetTruth( "MEMORY", "CGBank",		&CGBank,	CGBank );
	Ini->GetTruth( "MEMORY", "UseExtRam",	&UseExtRam,	UseExtRam );
	Ini->GetInt(   "MEMORY", "M1Wait",		&M1Wait,	M1Wait );
	Ini->GetTruth( "MEMORY", "UseSoldier",	&UseSol,	UseSol );
	//Ini->GetInt(   "MEMORY", "SolBank",		&SolBank,	SolBank );
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( &dt[2], &strva[j*2], 2 );
				IntRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// ORAM
	for( int i=0; i<(int)IEXTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "ExtRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( &dt[2], &strva[j*2], 2 );
				ExtRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// gROM
	Ini->GetString( "MEMORY", "FilePath", strfp, "" );
	if( *strfp ){
		if( !MountExtRom( strfp ) ) UnmountExtRom();
	}else
		UnmountExtRom();

	// -------- fixed by Windy 2010/3/25  ̕œǂݍނƁAUnmountExtRom()ŁAlɖ߂Ă܂̂ŁAœǂݍ
	Ini->GetInt(   "MEMORY", "SolBank",		&SolBank,	SolBank );

	SetMB_60();
	SetCGBank( CGBank );
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( SolBank );	// oN
	// -------------------------------------------------------------
	
	return TRUE;
}

BOOL MEM62::DokoLoad( cIni *Ini )
{
	int st;
	char strfp[PATH_MAX];
	
	if( !Ini ) return FALSE;
	
	Ini->GetTruth( "MEMORY", "CGBank",		&CGBank,	CGBank );
	Ini->GetTruth( "MEMORY", "UseExtRam",	&UseExtRam,	UseExtRam );
	Ini->GetTruth( "MEMORY", "cgrom",		&cgrom,		cgrom );
	Ini->GetTruth( "MEMORY", "kj_rom",		&kj_rom,	kj_rom );
	Ini->GetTruth( "MEMORY", "kj_LR",		&kj_LR,		kj_LR );
//	Ini->GetInt(   "MEMORY", "Rf0",			&st,		Rf0 );	Rf0 = st;
//	Ini->GetInt(   "MEMORY", "Rf1",			&st,		Rf1 );	Rf1 = st;
//	Ini->GetInt(   "MEMORY", "Rf2",			&st,		Rf2 );	Rf2 = st;
	Ini->GetInt(   "MEMORY", "M1Wait",		&M1Wait,	M1Wait );
	Ini->GetTruth( "MEMORY", "UseSoldier",	&UseSol,	UseSol );
//	Ini->GetInt(   "MEMORY", "SolBank",		&SolBank,	SolBank );
	

	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( dt, &strva[j*2], 2 );
				IntRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// ORAM
	for( int i=0; i<(int)IEXTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "ExtRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( dt, &strva[j*2], 2 );
				ExtRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// gROM
	Ini->GetString( "MEMORY", "FilePath", strfp, "" );
	if( *strfp ){
		if( !MountExtRom( strfp ) ) UnmountExtRom();
	}else
		UnmountExtRom();
	
	// -------- fixed by Windy 2010/3/25  ̕œǂݍނƁAUnmountExtRom()ŁAlɖ߂Ă܂̂ŁAœǂݍ
	Ini->GetInt(   "MEMORY", "Rf0",			&st,		Rf0 );	Rf0 = st;
	Ini->GetInt(   "MEMORY", "Rf1",			&st,		Rf1 );	Rf1 = st;
	Ini->GetInt(   "MEMORY", "Rf2",			&st,		Rf2 );	Rf2 = st;
	Ini->GetInt(   "MEMORY", "SolBank",		&SolBank,	SolBank );
	
	SetMB_R( Rf0, Rf1 );
	SetMB_W( Rf2 );
	SetCGBank( CGBank );
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( SolBank );	// oN
	// -------------------------------------------------------------
	
	return TRUE;
}


#ifdef USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
////////////////////////////////////////////////////////////////
// fobOp֐
////////////////////////////////////////////////////////////////
char *MEM6::GetReadMemBlk( int blk )
{
	return Rm_blk[blk]->GetName();
}

char *MEM6::GetWriteMemBlk( int blk )
{
	return Wm_blk[blk]->GetName();
}
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //



////////////////////////////////////////////////////////////////
//  device description
////////////////////////////////////////////////////////////////
const Device::Descriptor MEM60::descriptor = {
	MEM60::indef, MEM60::outdef
};

const Device::OutFuncPtr MEM60::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &MEM60::Out7FH )
};

const Device::InFuncPtr MEM60::indef[] = {};

const Device::Descriptor MEM62::descriptor = {
	MEM62::indef, MEM62::outdef
};

const Device::OutFuncPtr MEM62::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &MEM62::Out7FH ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC2H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC3H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF0H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF2H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF3H )
};

const Device::InFuncPtr MEM62::indef[] = {
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF0H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF1H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF2H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF3H )
};
