#ifndef MEMORY_H_INCLUDED
#define MEMORY_H_INCLUDED

#include "pc6001v.h"
#include "device.h"
#include "ini.h"


// ubN
#define MAXRMB	(16)
#define MAXWMB	(16)

////////////////////////////////////////////////////////////////
// NX`
////////////////////////////////////////////////////////////////
// ubNNX
class MemBlock {
protected:
	char Name[33];				// ubN
	BYTE *MB;					// ubNւ̃|C^
	int RWait;					// ANZXEFCg(Ǎ)
	int WWait;					// ANZXEFCg()
	BOOL WPt;					// CgveNgtO
	
public:
	MemBlock();									// RXgN^
	~MemBlock();								// fXgN^
	
	void SetMemory( const char *, BYTE *, int, int );	// ubNݒ
	void SetWait( int, int );					// ANZXEFCgݒ
	int GetWait();								// ANZXEFCg擾
	void SetProtect( BOOL );					// CgveNgݒ
	BOOL GetProtect();							// CgveNg擾
	
	char *GetName(){ return Name; }				// ubN擾
	
	BYTE Read( WORD, int * = NULL );			// [h
	void Write( WORD, BYTE, int * = NULL );		// Cg
};


class MEM6 : public IDoko {
protected:
	// \
	typedef struct{
		const char *FileName;	// t@C
		DWORD Size;				// TCY
		DWORD Crc;				// CRC32
		BYTE Init;				// f[^
		int WaitR;				// ANZXEFCg(Read)
		int WaitW;				// ANZXEFCg(Write)
	} MEMINFO;
	
	// 
	static const MEMINFO IEXTROM;
	static const MEMINFO IEMPTROM;
	static const MEMINFO IEMPTRAM;
	static const MEMINFO IEXTRAM;
	
	BOOL CGBank;				// CG ROM BANK TRUE:L FALSE:
	BOOL UseExtRom;				// gROM     TRUE:L FALSE:
	BOOL UseExtRam;				// gRAM     TRUE:L FALSE:
	
	BYTE *MainRom;				// BASIC ROM	60:16KB 62,66:32KB
	BYTE *ExtRom;				// g ROM		60:16KB 62,66:16KB
	BYTE *CGRom1;				// CG ROM1		60: 4KB 62,66: 8KB
	BYTE *IntRam;				//  RAM		60:16KB 62,66:64KB
	BYTE *ExtRam;				// O RAM		60:16KB 62,66:16KB
	BYTE *EmptyRom;				// ROM	60: 8KB 62,66: 8KB
	BYTE *EmptyRam;				// RAM	60: 8KB 62,66: 8KB
	
	MemBlock RomB[MAXRMB];		// ROMubN
	MemBlock RamB[MAXWMB];		// RAMubN
	
	MemBlock *Rm_blk[8];		// [hubN|C^(8KB*8)
	MemBlock *Wm_blk[8];		// CgubN|C^(8KB*8)
	
	BYTE ram_sel[4];			// RAM,ORAM Ip(16KB*4)
	
	char FilePath[PATH_MAX];	// gROMt@CtpX
	int M1Wait;					// M1EFCg
	BOOL EnableChkCRC;			// CRC`FbNL?
	
	
	// m̃J[gbW ------------------------------------------
	static const MEMINFO ISOLRAM;	// RAM
	BOOL UseSol;					// TRUE:L FALSE:
	int SolBank;					// ROMoN(0-15)
	void SetSolBank( BYTE );		// ROMoNݒ
	// -------------------------------------------------------------
	
	
	BOOL AllocMemory( BYTE **, const MEMINFO *, char * );	// mۂROMt@CǍ
	BOOL InitCommon( BOOL, BOOL );	// ʕ
	
public:
	MEM6( BOOL );							// RXgN^
	virtual ~MEM6();						// fXgN^
	
	virtual BOOL Init( char *, BOOL, BOOL ) = 0;	// 
	virtual void Patch(){};					// pb`
	virtual void Reset(){};					// Zbg
	
	BYTE Fetch( WORD, int * = NULL );		// tFb`(M1)
	BYTE Read( WORD, int * = NULL );		// [h
	void Write( WORD, BYTE, int * = NULL );	// Cg
	
	BOOL MountExtRom( char * );				// gROM }Eg
	void UnmountExtRom();					// gROM A}Eg
	char *GetFile();						// gROMt@CpX擾
	
	// 8255o͊֘A֐
	virtual void SetCGBank( BOOL );			// CG ROM BANK ؂ւ
	
	// ڃANZX֐
	virtual BYTE ReadMainRom( WORD addr ){ return MainRom[addr&0x3fff]; }
	virtual BYTE ReadMainRam( WORD addr ){ return IntRam[addr&0x3fff]; }
	virtual BYTE ReadExtRom ( WORD addr ){ return ExtRom ? ExtRom[addr&0x3fff] : 0xff; }
	virtual BYTE ReadCGrom1 ( WORD addr ){ return CGRom1[addr&0x1fff]; }
	virtual BYTE ReadCGrom2 ( WORD )     { return 0xff; }
	
	#ifdef USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	virtual char *GetReadMemBlk( int );
	virtual char *GetWriteMemBlk( int );
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
};


class MEM60 : public MEM6, public Device {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO IINTRAM;
	
	void SetMB_60();					// ubNݒ(60̂)
	
	// foCX`
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/OANZX֐
	inline void Out7FH( int, BYTE data ){ SetSolBank( data ); }
	
public:
	MEM60( const ID&, BOOL );			// RXgN^
	~MEM60();							// fXgN^
	
	BOOL Init( char *, BOOL, BOOL );	// 
	void Patch();						// pb`
	void Reset();						// Zbg
	
	// ڃANZX֐
	BYTE ReadMainRam( WORD addr ){ return addr<0x4000 ? ExtRam[addr&0x3fff] : IntRam[addr&0x3fff]; }
	
	// foCXID
	enum IDOut{ out7FH=0 };
	enum IDIn {};
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ǂłSAVE
	BOOL DokoLoad( cIni * );	// ǂłLOAD
	// ------------------------------------------
};


class MEM62 : public MEM6, public Device {
protected:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IINTRAM;
	
	BYTE *CGRom2;		// CG ROM2		60: -KB 62,66: 8KB
	BYTE *KanjiRom;		//  ROM		60: -KB 62,66:32KB
	BYTE *VoiceRom;		//  ROM	60: -KB 62,66:16KB
	
	BOOL cgrom;			// CG ROM Ip TRUE:N60[h FALSE:N60m[h
	BOOL kj_rom;		// ROM,ROM Ip TRUE:ROM FALSE:ROM
	BOOL kj_LR;			// ROM EIp TRUE:E FALSE:
	
	BYTE Rf0;			// Rg[WX^
	BYTE Rf1;			// Rg[WX^
	BYTE Rf2;			// Rg[WX^
	
	void SetMB_R( BYTE, BYTE );			// [h̃ubNw(62,66̂)
	void SetMB_W( BYTE );				// Cg̃ubNw(62,66̂)
	void SetWait( BYTE );				// ANZXEFCgݒ
	BYTE GetWait();						// ANZXEFCg擾
	
	void SelectCGrom( int );			// CG ROM I(62,66̂)
	void SelectKanjiRom( BYTE );		// ROM  ROM ؂ւ(62,66̂)
	
	// foCX`
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/OANZX֐
	inline void Out7FH( int, BYTE data ){ SetSolBank( data ); }
	inline void OutC1H( int, BYTE data ){ SelectCGrom( (data>>1)&1 ); }
	inline void OutC2H( int, BYTE data ){ SelectKanjiRom( data ); }
	inline void OutC3H( int, BYTE data ){}
	inline void OutF0H( int, BYTE data ){ SetMB_R( data, Rf1 ); }
	inline void OutF1H( int, BYTE data ){ SetMB_R( Rf0, data ); }
	inline void OutF2H( int, BYTE data ){ SetMB_W( data ); }
	inline void OutF3H( int, BYTE data ){ SetWait( data ); }
	
	inline BYTE InF0H( int ){ return Rf0; }
	inline BYTE InF1H( int ){ return Rf1; }
	inline BYTE InF2H( int ){ return Rf2; }
	inline BYTE InF3H( int ){ return GetWait() | 0x1f; }
	
public:
	MEM62( const ID&, BOOL );			// RXgN^
	virtual ~MEM62();					// fXgN^
	
	virtual BOOL Init( char *, BOOL, BOOL );	// 
	void Patch();						// pb`
	void Reset();						// Zbg
	
	// 8255o͊֘A֐
	void SetCGBank( BOOL );				// CG ROM BANK ؂ւ
	
	// ڃANZX֐
	BYTE ReadMainRom( WORD addr ) { return MainRom[addr&0x7fff]; }
	BYTE ReadMainRam( WORD addr ) { return IntRam[addr&0xffff]; }
	BYTE ReadCGrom2( WORD addr )  { return CGRom2[addr&0x1fff]; }
	BYTE ReadKanjiRom( WORD addr ){ return KanjiRom[addr&0x7fff]; }
	BYTE ReadVoiceRom( WORD addr ){ return VoiceRom[addr&0x3fff]; }
	
	// foCXID
	enum IDOut{ out7FH=0, outC1H, outC2H, outC3H, outF0H, outF1H, outF2H, outF3H };
	enum IDIn {  inF0H=0, inF1H,  inF2H,  inF3H };
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ǂłSAVE
	BOOL DokoLoad( cIni * );	// ǂłLOAD
	// ------------------------------------------
};


class MEM66 : public MEM62 {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IINTRAM;
	
	void SelectKanjiRom( BYTE );		// ROM  ROM ؂ւ(62,66̂)
	
public:
	MEM66( const ID&, BOOL );			// RXgN^
	~MEM66();							// fXgN^
	
	BOOL Init( char *, BOOL, BOOL );	// 
	void Patch();						// pb`
};


#endif	// MEMORY_H_INCLUDED
