#include <new>
#include <math.h>
#include <memory.h>

#include <SDL_endian.h>

#include "sound.h"
#include "error.h"


#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#define	AUDIOFORMAT	AUDIO_S16MSB	// 16rbg
#else
#define	AUDIOFORMAT	AUDIO_S16		// 16rbg
#endif


////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
cRing::cRing( void )
{
	Buffer = NULL;
	Size = Wpt = Rpt = Num = 0;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
cRing::~cRing( void )
{
	if( Buffer ) delete[] Buffer;
}


////////////////////////////////////////////////////////////////
// obt@
//
// :	size		Tv
// Ԓl:	BOOL		TRUE: FALSE:s
////////////////////////////////////////////////////////////////
BOOL cRing::InitBuffer( int size )
{
	PRINTD1( SND_LOG, "[cRing][Init] Size:%d\n", size );
	
	if( Buffer ) delete [] Buffer;
	
	Size = size*MULTI;	// ۂ̃obt@TCYMULTI{
	Wpt = Rpt = Num = 0;
	
	try{
		Buffer = new int[Size];
		memset( Buffer, 0, Size );
	}
	// new Ɏsꍇ
	catch( std::bad_alloc ){
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// Ǎ
//
// :	idx		CfbNX(-1:1f[^ǂŃ|C^i߂)
//								(0-:idx1f[^ǂ |C^i߂Ȃ)
// Ԓl:	int		f[^
////////////////////////////////////////////////////////////////
int cRing::Get( int idx )
{
	if( Num ){
		int data = 0;
		SCritical::Lock();
		if( idx < 0 ){
			data = Buffer[Rpt++];
			if( Rpt == Size ) Rpt = 0;
			Num--;
		}else{
			if( idx < Size ){
				int pt = Rpt + idx;
				if( pt >= Size ) pt -= Size;
				data = Buffer[pt];
			}
		}
		SCritical::UnLock();
		return data;
	}else{
		return 0;
	}
}


////////////////////////////////////////////////////////////////
// 
//
// :	data		f[^
// Ԓl:	BOOL		TRUE: FALSE:s(obt@ς)
////////////////////////////////////////////////////////////////
BOOL cRing::Put( int data )
{
	if( Num < Size ){
		SCritical::Lock();
		Buffer[Wpt++] = data;
		if( Wpt == Size ) Wpt = 0;
		Num++;
		SCritical::UnLock();
		return TRUE;
	}else
		return FALSE;
}


////////////////////////////////////////////////////////////////
// ǃf[^擾
//
// :	Ȃ
// Ԓl:	int			ǃTv
////////////////////////////////////////////////////////////////
int cRing::ReadySize( void )
{
	return Num;
}


////////////////////////////////////////////////////////////////
// cobt@擾
//
// :	Ȃ
// Ԓl:	int			cobt@Tv
////////////////////////////////////////////////////////////////
int cRing::FreeSize( void )
{
	return Size - Num;
}


////////////////////////////////////////////////////////////////
// obt@TCY擾
//
// :	Ȃ
// Ԓl:	int			Sobt@Tv
////////////////////////////////////////////////////////////////
int cRing::GetSize( void )
{
	return Size;
}




////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
SndDev::SndDev( void )
{
	Volume  = 0;
	LPF_Mem = 0;
	LPF_k   = 0x8000;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
SndDev::~SndDev( void ){}


////////////////////////////////////////////////////////////////
// TvO[g擾
// :	Ȃ
// Ԓl:	int		TvO[g
////////////////////////////////////////////////////////////////
int SndDev::GetSampleRate( void )
{
	return DEFAULT_SAMPLE_RATE;
}


////////////////////////////////////////////////////////////////
// [pXtB^ JbgItgݒ
//
// :	fc		JbgItg
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SndDev::SetLPF( int fc )
{
	if( fc == 0 )
		LPF_k = 0x8000;
	else
		LPF_k = (int)((2.0*M_PI*(double)fc*(double)0x8000)/(double)GetSampleRate());
}


////////////////////////////////////////////////////////////////
// [pXtB^
//
// [Ql] ̃fBW^tB^
// http://kmkz.jp/mtm/mag/lab/digitalfilter.htm
//
// :	src		f[^
// Ԓl:	int		tB^Kp̃f[^
////////////////////////////////////////////////////////////////
int SndDev::LPF( int src )
{
	LPF_Mem += ( ( src - LPF_Mem ) * LPF_k ) / 0x8000;
	return LPF_Mem;
}


////////////////////////////////////////////////////////////////
// Ǎ(I[o[Ch)
//
// :	idx		CfbNX(-1:1f[^ǂŃ|C^i߂)
//								(0-:idx1f[^ǂ |C^i߂Ȃ)
// Ԓl:	int		f[^
////////////////////////////////////////////////////////////////
int SndDev::Get( int idx )
{
	return LPF( cRing::Get( idx ) );
}


////////////////////////////////////////////////////////////////
// ʐݒ
//
// :	vol		(0-100)
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SndDev::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}


////////////////////////////////////////////////////////////////
// Xg[XV
//
// :	samples	VTv
// Ԓl:	int		XVTv
////////////////////////////////////////////////////////////////
int SndDev::SoundUpdate( int samples )
{
	return 0;
}


////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
SND6::SND6( void )
{
	memset( &ASpec, 0, sizeof(SDL_AudioSpec) );
	for( int i=0; i<MAXSTREAM; i++ ) RB[i] = NULL;
	Volume = 0;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
SND6::~SND6( void )
{
	Pause();
	// I[fBIfoCX
	SDL_CloseAudio();
}


////////////////////////////////////////////////////////////////
// 
//
// :	p6			gւ̃IuWFNg|C^
//			callback	R[obN֐ւ̃|C^
//			rate		TvO[g
//			size		obt@TCY({)
// Ԓl:	BOOL		TRUE: FALSE:s
////////////////////////////////////////////////////////////////
BOOL SND6::Init( void *p6, void (*callback)(void *, Uint8 *, int ), int rate, int size )
{
	PRINTD( SND_LOG, "[SND6][Init]\n" );
	
	ASpec.freq     = rate;							// TvO[g
	ASpec.format   = AUDIOFORMAT;					// tH[}bg
	ASpec.channels = 1;								// m
	ASpec.samples  = rate * size / VSYNC_HZ;		// obt@TCY(Tv)
	ASpec.callback = callback;						// R[obN֐̎w
	ASpec.userdata = p6;							// R[obN֐Ɏg̃IuWFNg|C^n
	
	for( int i=0; i<MAXSTREAM; i++ ) RB[i] = NULL;
	
	Volume = DEFAULT_MASTERVOL;
	
	// obt@
	if( !cRing::InitBuffer( ASpec.samples ) ) return FALSE;
	
	// I[fBIfoCXJ
	if( SDL_OpenAudio( &ASpec, NULL ) < 0 ) return FALSE;
	
	PRINTD1( SND_LOG, " SampleRate : %d\n",     ASpec.freq );
	PRINTD1( SND_LOG, " Format     : 0x%04X\n", ASpec.format );
	PRINTD1( SND_LOG, " Channels   : %d\n",     ASpec.channels );
	PRINTD1( SND_LOG, " Size       : %d\n",     ASpec.size );
	PRINTD1( SND_LOG, " Silence    : %d\n",     ASpec.silence );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// Xg[ڑ
//
// :	SndDev *	obt@|C^
// Ԓl:	BOOL		TRUE: FALSE:s
////////////////////////////////////////////////////////////////
BOOL SND6::ConnectStream( SndDev *buf )
{
	PRINTD( SND_LOG, "[SND6][ConnectStream]\n" );
	
	if( !buf ) return FALSE;
	
	for( int i=0; i<MAXSTREAM; i++ ){
		if( !RB[i] ){
			// obt@
			if( !buf->InitBuffer( ASpec.samples ) ) return FALSE;
			RB[i] = buf;
			return TRUE;
		}
	}
	
	return FALSE;
}


////////////////////////////////////////////////////////////////
// Đ
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SND6::Play( void )
{
	PRINTD( SND_LOG, "[SND6][Play]\n" );
	
	SDL_PauseAudio( 0 );
}


////////////////////////////////////////////////////////////////
// ~
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SND6::Pause( void )
{
	PRINTD( SND_LOG, "[SND6][Pause]\n" );
	
	SDL_PauseAudio( 1 );
}


////////////////////////////////////////////////////////////////
// TvO[g擾
//
// :	Ȃ
// Ԓl:	int		TvO[g
////////////////////////////////////////////////////////////////
int SND6::GetSampleRate( void )
{
	return ASpec.freq;
}


////////////////////////////////////////////////////////////////
// }X^[ʐݒ
//
// :	vol		(0-100)
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SND6::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}


////////////////////////////////////////////////////////////////
// TEhOXV֐
//
// :	samples		Tv
//			exbuf		Oobt@|C^
// Ԓl:	int			XVTv
////////////////////////////////////////////////////////////////
int SND6::PreUpdate( int samples, cRing *exbuf )
{
	int exsam = 0;
	
	PRINTD3( SND_LOG,"PreUpdate %d %d %d \n", RB[0]->ReadySize(), RB[1]->ReadySize(), RB[2]->ReadySize() );
	
	for( int i=0; i<MAXSTREAM; i++ )
		if( RB[i] ) exsam = min( max( exsam, RB[i]->ReadySize() ), samples );
	
	for( int i=0; i<exsam; i++ ){
		int dat = 0;
		for( int j=0; j<MAXSTREAM; j++ )
			if( RB[j] ) dat += RB[j]->Get();
		
		dat = ( dat * Volume ) / 100;
		dat = min( max( dat, -32768 ), 32767 );
		
		cRing::Put( (Sint16)dat );
		// Oobt@݂Ώ
		if( exbuf ) exbuf->Put( (Sint16)dat );
	}
	
	return exsam;
}


////////////////////////////////////////////////////////////////
// TEhXV֐
//
// :	stream		Xg[݃obt@ւ̃|C^
//			samples		Tv
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SND6::Update( Uint8 *stream, int samples )
{
	Sint16 *str = (Sint16 *)stream;
	
	PRINTD3( SND_LOG, "[SND6][Update] Stream:%08X Samples:%d / %d\n", (int)stream, samples, cRing::ReadySize() );
	
	for( int i=0; i<samples; i++ ){
		*(str++) = (Sint16)cRing::Get();
	}
}
